package arule;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.TreeSet;

import miner.util.ObjectTuple;
import util.ArrayUtil;
import util.FormatUtil;

public class AssociationRule implements Comparable<AssociationRule>{
	
	private ObjectTuple premise;
	private ObjectTuple conclusion;
	
	private int[] abc = new int[4];
	private int dataSize;
	
	public AssociationRule(ObjectTuple premise, ObjectTuple conclusion)
	{
		this.premise 	= premise;
		this.conclusion = conclusion;
	}
	
	public static TreeSet<AssociationRule> extract(ArrayList<ArrayList<ObjectTuple>> freqP_LL, ArrayList<int[]> rData)
	{
		int[][] rankings = new int[rData.size()][];
		rData.toArray(rankings);
		
		return extract( freqP_LL, rankings);
	}
	
	public static TreeSet<AssociationRule> extract(ArrayList<ArrayList<ObjectTuple>> freqP_LL, int[][] rData)
	{
		ArrayList<ObjectTuple> oData = new ArrayList<>();
		
		for(int[] iA: rData)
		{
			oData.add(ObjectTuple.rankingToObjectTuple(iA));
		}
		
		TreeSet<AssociationRule> ret_S = new TreeSet<AssociationRule>();
		
		AssociationRule ar = null;
		
		for(int i=0; i<freqP_LL.size(); i++)
		for(int j=0; j<freqP_LL.get(i).size(); j++)
		{
			for(int k=0; k<freqP_LL.size(); k++)
			for(int l=0; l<freqP_LL.get(k).size(); l++)
			{
				if(freqP_LL.get(i).get(j).isConcordant(freqP_LL.get(k).get(l)))
				if(!freqP_LL.get(i).get(j).contains(freqP_LL.get(k).get(l)))
				{
					ar = new AssociationRule(freqP_LL.get(i).get(j),freqP_LL.get(k).get(l));
					
					ar.setABC(compABC(ar, oData));
					ar.dataSize = oData.size();
					ret_S.add(ar);
				}
			}
		}		
		return ret_S;
	}
	
	public static TreeSet<AssociationRule> extract2(ArrayList<ArrayList<ObjectTuple>> freqP_LL, ArrayList<int[]> rData, double freq, double conf, double interest, double redu)
	{
		int[][] rankings = new int[rData.size()][];
		rData.toArray(rankings);
		
		return extract2( freqP_LL, rankings, freq, conf, interest, redu);
	}
	
	public static TreeSet<AssociationRule> extract2(ArrayList<ArrayList<ObjectTuple>> freqP_LL, int[][] rData, double freq, double conf, double interest, double redu)
	{
		ArrayList<ObjectTuple> oData = new ArrayList<>();
		
		for(int[] iA: rData)
		{
			oData.add(ObjectTuple.rankingToObjectTuple(iA));
		}
		
		TreeSet<AssociationRule> ret_S = new TreeSet<AssociationRule>();
		
		AssociationRule ar = null;
		
		for(int i=0; i<freqP_LL.size(); i++)
		for(int j=0; j<freqP_LL.get(i).size(); j++)
		{
			for(int k=0; k<freqP_LL.size(); k++)
			for(int l=0; l<freqP_LL.get(k).size(); l++)
			{
				if(!freqP_LL.get(i).get(j).contains(freqP_LL.get(k).get(l)))
				{
					ar = new AssociationRule(freqP_LL.get(i).get(j),freqP_LL.get(k).get(l));
										
					ar.setABC(compABC(ar, oData));
					ar.dataSize = rData.length;
									
					if(freq <= ar.abc[0]/(double)rData.length
							&& conf <= ar.confidence()
							&& interest <= ar.interest()
							&& redu >= ar.redundancy())
					{
						ar = ar.reduceRedundancy(oData);
						ret_S.add(ar);
					}
				}
			}
		}		
		return ret_S;
	}
	
//	public static ArrayList<ObjectTuple[]> splitTuple(ObjectTuple oTuple)
//	{
//		ArrayList<ObjectTuple[]> ret_L = new ArrayList<>();
//		
//		int size = oTuple.size();
//		
//		for(int i=0; i<size; i++)
//		{
//			for(int j=0; j<size; j++)
//			{
//				if(i != j)
//				{
//					
//				}
//			}
//		}
//		
//		
//		return ret_L;
//	}
	
//	private static void splitTuple_R(ArrayList<ObjectTuple[]> ret_L, ObjectTuple oTuple)
	
	
	public static int[] compABC(AssociationRule rar, ArrayList<ObjectTuple> oData)
	{
		int[] ret_A = new int[4];
		
		int[] sABC;
		
		ret_A[3] = oData.size();
		
		for(ObjectTuple ot: oData)
		{
			sABC = compSingleABC(rar, ot);
//System.out.println(rar.getPremise().toString() + " " + rar.getConclusion().toString() + " odp " + ot.toString() + " " + sABC[0] + " " + sABC[1] + " " + sABC[2]);			
			ret_A[0] = ret_A[0] + sABC[0];
			ret_A[1] = ret_A[1] + sABC[1];
			ret_A[2] = ret_A[2] + sABC[2];
		}
		
		return ret_A;
	}
	
	public static int[] compSingleABC(AssociationRule rar, ObjectTuple element)
	{	
		if(element.contains(rar.getPremise()))
		{
			if(element.contains(rar.getConclusion()))
			{
				return new int[]{1, 1, 1};
			}
			else
			{
				return new int[]{0, 1, 0};
			}
		}
		else if(element.contains(rar.getConclusion()))
		{
			return new int[]{0, 0, 1};
		}
		
		return new int[]{0, 0, 0};
	}
	
	public AssociationRule reduceRedundancy(ArrayList<ObjectTuple> oData)
//	public AssociationRule reduceRedundancy()
	{
		HashSet<Integer> prem_S = new HashSet<>();
		
		ObjectTuple premN = new ObjectTuple();
		ObjectTuple concN = new ObjectTuple();
		
		for(int i=0; i<premise.size(); i++)
		{
			prem_S.add(premise.get(i));
			premN.add(new Integer(premise.get(i)));
		}
		
		if(!(prem_S.contains(conclusion.get(0)) && prem_S.contains(conclusion.get(1))))
		{
			concN.add(conclusion.get(0));
		}
		for(int i=1; i<this.conclusion.size()-1; i++)
		{
//System.out.println(conclusion.get(i) + " " + prem_S.contains(conclusion.get(i)) + " " 
//									+ prem_S.contains(conclusion.get(i)) + " " 
//									+ prem_S.contains(conclusion.get(i+1)));
			if(!(prem_S.contains(conclusion.get(i-1)) && prem_S.contains(conclusion.get(i)) && prem_S.contains(conclusion.get(i+1))))
			{
				concN.add(conclusion.get(i));
			}
		}
		if(!(prem_S.contains(conclusion.get(this.conclusion.size()-2)) && prem_S.contains(conclusion.get(this.conclusion.size()-1))))
		{
			concN.add(conclusion.get(this.conclusion.size()-1));
		}
	
		AssociationRule ar = new AssociationRule(premN,concN);
		ar.setABC(compABC(ar, oData));
		ar.dataSize = oData.size();
System.out.println(this.toString() + " " + ar.toString());
//		
		return ar;	
	}
	
	public ObjectTuple getPremise()
	{
		return this.premise;
	}
	
	public ObjectTuple getConclusion()
	{
		return this.conclusion;
	}
	public int[] getABC()
	{
		return this.abc;
	}
	
	public void setABC(int[] abc)
	{
		this.abc = abc;
	}
	
	public double confidence()
	{
		return this.abc[0]/(double)this.abc[1];
	}
	
	public double interest()
	{
		return this.confidence() - (this.abc[2]/(double)this.abc[3]);
	}
	
	public double redundancy()
	{
		int numInters = this.premise.intersectObjects(this.conclusion).size();
		
		return numInters * (numInters-1) /(double)(premise.size() * (premise.size()-1) +  conclusion.size() * (conclusion.size()-1) - numInters * (numInters-1));
	}
	
	


	@Override
	public int compareTo(AssociationRule ar) {
		
		//confidence
//		if(this.confidence() > ar.confidence()) return -1;
		
		//interest
		if(this.interest() > ar.interest()) return -1;
		
		return 1;
	}
	
	@Override
	public String toString()
	{
		StringBuffer sb = new StringBuffer();
		
		sb.append(premise.toString());
		sb.append(" -> ");
		sb.append(conclusion.toString());		
		sb.append("  " + "(" + abc[0] + "," + abc[1] + "," + abc[2] + ")");
		sb.append(" Support: " + "(" + FormatUtil.cutDecimal(abc[0]/(double)dataSize,3) + ")");
		sb.append(" Conf: " + "(" + FormatUtil.cutDecimal(this.confidence(),3) + ")");
		sb.append(" Intrest: " + "(" + FormatUtil.cutDecimal(this.interest(),3) + ")");
		sb.append(" redundancy: " + "(" + FormatUtil.cutDecimal(this.redundancy(),3) + ")");
		
		return sb.toString();
	}
	
	public String toLatex()
	{
		StringBuffer sb = new StringBuffer();
		
		sb.append("$");
		sb.append(premise.toString());
		sb.append(" \rightarrow ");
		sb.append(conclusion.toString());
		sb.append("$");
		sb.append("  " + "(" + abc[0] + "," + abc[1] + "," + abc[2] + ")");
		sb.append(" Conf: " + "(" + abc[0]/(double)abc[1] + ")");
		sb.append(" Intrest: " + "(" + (abc[0]/(double)abc[1] - (this.abc[2]/(double)this.abc[3])) + ")\n");
		
		return sb.toString();
	}

}
