package miner.tesma;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map.Entry;
import java.util.Scanner;
import java.util.TreeSet;

import miner.RankingMiner;
import miner.util.DataMaster;
import miner.util.ObjectTuple;
import miner.util.TransactionList;
import util.MemoryMonitor;

public class TESMA2 implements RankingMiner{

	private DataMaster dMaster;
	private ArrayList<TreeSet<ObjectTuple>> frqP_LL;
	
	private int absoluteT;
//	private TransactionList intersection = null;
	private TransactionList[] intersection_A;
	
	MemoryMonitor memMon;
	
	ObjectTuple otf;
	
//	Scanner keyboard = new Scanner(System.in); //Test
	
	//#########################################################################
	// Constructor
	//#########################################################################
	
	public TESMA2(int[][] rankings)
	{
		dMaster = new DataMaster(rankings);
		this.intersection_A = new TransactionList[rankings[0].length];
		initInterList(rankings.length);
	}
	
	public TESMA2(ArrayList<int[]> rankings)
	{
		dMaster = new DataMaster(rankings);
		this.intersection_A = new TransactionList[rankings.get(0).length];
		initInterList(rankings.size());
	}
	
	private void initInterList(int size)
	{
		for(int i=0; i<intersection_A.length; i++)
		{
			intersection_A[i] = new TransactionList(size);
		}
	}
	
	//#########################################################################
	// Override RankingMiner
	//#########################################################################
	
	
	@Override
	public void mine(double t) {

		//##### MM #########
//		System.runFinalization();
//		memMon = new MemoryMonitor();
//		memMon.setStartMemory();
		//##################	
		absoluteT = (int)Math.ceil((t * dMaster.size()));

		frqP_LL = new ArrayList<TreeSet<ObjectTuple>>();
		
		//Generate one layer for every object
		for(int i=0; i<dMaster.numRanks(); i++)
		{
			frqP_LL.add(new TreeSet<ObjectTuple>());
		}

		dMaster.compFrq2Tuple(t);
		
		ObjectTuple ot = null;
		
		//Fill first layer (2-rankings)
		for(Entry<Integer, HashMap<Integer,TransactionList>> e: dMaster.getFrequentTuple().entrySet())
		{
			for(Entry<Integer,TransactionList> ep: e.getValue().entrySet())
			{
				ot = new ObjectTuple();
				ot.add(e.getKey());
				ot.add(ep.getKey());
				ot.setAbsFrq(ep.getValue().count());
				frqP_LL.get(0).add(ot);
				
				//########MM#######				
//				memMon.updateMemory();;
				//#################
				depthFirst(ot, 0, ep.getValue());
			}
		}		

		for(int i= frqP_LL.size(); i> 0; i--)
		{
			if(frqP_LL.get(i-1).size() == 0)
				frqP_LL.remove(frqP_LL.size()-1);
		}
	}
	
	public void depthFirst(ObjectTuple ot, int k, TransactionList tl)
	{
		if(dMaster.getFrequentTuple().containsKey(ot.getLast()))
		{
			for(Entry<Integer,TransactionList> ep: dMaster.getFrequentTuple().get(ot.getLast()).entrySet())
			{
//				intersection = ep.getValue().AND(tl);
//				int iCount = intersection.count();
				
				intersection_A[k].assimilate(ep.getValue());
				TransactionList.AND(intersection_A[k],tl);				
				int iCount = intersection_A[k].count();
				
				if(iCount >= absoluteT)
				{
					otf = new ObjectTuple(ot);
					otf.add(ep.getKey());
					otf.setAbsFrq(iCount);
					frqP_LL.get(k+1).add(otf);
					
					//########MM#######
//					memMon.updateMemory();
					//#################
//					depthFirst(otf, k+1, intersection);
					depthFirst(otf, k+1, intersection_A[k]);
				}
			}
		}
	}
	
	public ArrayList<ArrayList<ObjectTuple>> getTree()
	{
		return null;
	}

	@Override
	public ArrayList<ArrayList<ObjectTuple>> getFrequent() 
	{
		ArrayList<ArrayList<ObjectTuple>> ret_LL = new ArrayList<>();
		
		for(int i=0; i<frqP_LL.size(); i++)
		{			
			ret_LL.add(new ArrayList<ObjectTuple>());
			
			for(ObjectTuple ot: frqP_LL.get(i))
			{
				ret_LL.get(i).add(ot);
			}
		}		
		return ret_LL;
	}
	
	@Override
	public ArrayList<ArrayList<ObjectTuple>> getClosed() 
	{
		ArrayList<ArrayList<ObjectTuple>> ret_LL = new ArrayList<>();
		ArrayList<ObjectTuple> ret_L = null;
		
		Iterator<ObjectTuple> it = null;
		for(TreeSet<ObjectTuple> ot_S: frqP_LL)
		{
			ret_L = new ArrayList<ObjectTuple>();
			ret_LL.add(ret_L);
			it = ot_S.iterator();
			while(it.hasNext())
			{
				ret_L.add((ObjectTuple)it.next());
			}
		}		
		ArrayList<ObjectTuple> remove_L = new ArrayList<>();			
		TreeSet<ObjectTuple> currTuple = null;
		
		ObjectTuple iot;
		ObjectTuple i2ot;
		Iterator<ObjectTuple> iter;
		Iterator<ObjectTuple> iter2;
		
		int compRes;
		boolean breakOuter;
		
		for(int i=1; i<frqP_LL.size(); i++)
		{
			for(ObjectTuple ot: frqP_LL.get(i))
			{
				currTuple = ot.getSubTuple();		
				iter = ret_LL.get(i-1).iterator();
				iter2 = currTuple.iterator();
				iot = null;
				i2ot = (ObjectTuple)iter2.next();

				while(iter.hasNext())
				{
					iot = (ObjectTuple)iter.next();
//System.out.println(iot.toString() + " vs. " + i2ot.toString() + " sub of " + ot.toString());

					compRes = iot.compareTo(i2ot);
					
					if(compRes < 0)
					{
						//mach garnichts
					}
					else if(compRes == 0)
					{
						if(iot.getAbsFrq() == ot.getAbsFrq())
						{
							remove_L.add(iot);							
						}
						if(iter2.hasNext()) i2ot = (ObjectTuple)iter2.next();
						else				break;
					}
					else
					{
						breakOuter = false;
						while(iter2.hasNext())
						{
							i2ot = (ObjectTuple)iter2.next();
//System.out.println(iot.toString() + " vs. " + i2ot.toString() + " sub of " + ot.toString());
							compRes = i2ot.compareTo(iot);
							
							if(compRes < 0)
							{
								//mach garnichts
							}
							else if(compRes == 0)
							{
								if(iot.getAbsFrq() == ot.getAbsFrq())
								{
									remove_L.add(iot);							
								}
								if(iter2.hasNext())
								{
									i2ot = (ObjectTuple)iter2.next();
									break;
								}
								else
								{
									breakOuter = true;
									break;
								}
							}
							else
							{
								break;
							}
						}
						if(breakOuter) break;
					}
				}
				ret_LL.get(i-1).removeAll(remove_L);
				remove_L.clear();
			}
		}
		return ret_LL;
	}


	public ArrayList<ArrayList<ObjectTuple>> getClosedV2() 
	{
		ArrayList<ArrayList<ObjectTuple>> ret_LL = new ArrayList<>();
		ArrayList<ObjectTuple> ret_L = null;
		
		Iterator<ObjectTuple> it = null;
		for(TreeSet<ObjectTuple> ot_S: frqP_LL)
		{
			ret_L = new ArrayList<ObjectTuple>();
			ret_LL.add(ret_L);
			it = ot_S.iterator();
			while(it.hasNext())
			{
				ret_L.add((ObjectTuple)it.next());
			}
		}		
		ArrayList<ObjectTuple> remove_L = new ArrayList<>();			
		TreeSet<ObjectTuple> currTuple = null;
		
		ObjectTuple iot;
		ObjectTuple i2ot;
		Iterator<ObjectTuple> iter;
		Iterator<ObjectTuple> iter2;
		
		for(int i=1; i<frqP_LL.size(); i++)
		{
			for(ObjectTuple ot: frqP_LL.get(i))
			{
				currTuple = ot.getSubTuple();		
				iter = frqP_LL.get(i-1).iterator();
				iter2 = currTuple.iterator();
				iot = null;
				i2ot = (ObjectTuple)iter2.next();

				while(iter.hasNext())
				{

					iot = (ObjectTuple)iter.next();

					if((iot).equals(i2ot))
					{
						if(iot.getAbsFrq() == ot.getAbsFrq())
						{

							remove_L.add(iot);							
						}
						if(iter2.hasNext()) i2ot = (ObjectTuple)iter2.next();
						else				break;
					}
				}
				ret_LL.get(i-1).removeAll(remove_L);
				remove_L.clear();
			}
		}
		return ret_LL;
	}
	
	public ArrayList<ArrayList<ObjectTuple>> getClosedBottomUp() 
	{
		ArrayList<ArrayList<ObjectTuple>> ret_LL = new ArrayList<>();
		ArrayList<ObjectTuple> ret_L = null;
		
		for(int i=0; i< frqP_LL.size(); i++)
		{
			ret_L = new ArrayList<ObjectTuple>();
			ret_LL.add(ret_L);
		}
		
		boolean otRaus = false;
		
		for(int i=0; i<frqP_LL.size()-1; i++)
		{		
			for(ObjectTuple otB: frqP_LL.get(i))
			{
				otRaus = false;
				
				for(ObjectTuple otT: frqP_LL.get(i+1))
				{					
					if(otB.getAbsFrq() == otT.getAbsFrq())
						if((otT).contains(otB))
						{
							otRaus = true;
							break;
						}
				}
				if(!otRaus)
				{
					ret_LL.get(i).add(otB);
				}
			}
		}
		
		for(ObjectTuple ot: frqP_LL.get(frqP_LL.size()-1))
		{
			ret_LL.get(frqP_LL.size()-1).add(ot);
		}
		
		
		return ret_LL;
	}

	@Override
	public ArrayList<ArrayList<ObjectTuple>> getMaximal()
	{		
		//TODO
		return null;
	}

	@Override
	public int getID() {
		
		return 102;
	}
	
	public long getUsedMemory()
	{
		return memMon.getUsedMemory();
	}

}

