package miner.util;

import java.util.ArrayList;

public class PairVector {
	
	private BitVector bv;
	private int n;
	
	public PairVector(PairVector pv)
	{
		this.bv = new BitVector(pv.bv);
		this.n = pv.n;
	}
	
	public PairVector(BitVector bv, int n)
	{
		this.bv = bv;
		this.n = n;
	}
	
	public PairVector(int n)
	{
		bv = new BitVector(n * n - 1);
		this.n = n;
	}
	
	public PairVector(int[] ranking)
	{
		bv = new BitVector(ranking.length * ranking.length - 1);
		n = ranking.length;
		
		int[] help = new int[ranking.length];
		int length = 0;
		
		for(int i=0; i<help.length; i++)
		{
			if(ranking[i] > 0)
			{
				help[ranking[i]-1] = i+1;
				length++;
			}
		}
	
		for(int i=0; i<length-1; i++)
		{
			for(int j=i+1; j<length; j++)
			{
				this.add(help[i],help[j]);
			}
		}
	}
	
	public PairVector AND(PairVector pv)
	{
		return new PairVector(bv.AND(pv.bv),n);
	}
	
	public ArrayList<int[]> getPairs()
	{
		ArrayList<int[]> ret_L = new ArrayList<>();
		
		int trans;
		int index;
		
		int b, c, p, o;
		
		for(int i=0; i<bv.list.length; i++)
		{
			trans = 31*i;
			
			b = bv.list[i];
			c = 0;
			
			while(b != 0)
			{
				c = b & (0 - b);
				
				index = trans + Integer.numberOfTrailingZeros(c) + 1;
				p = Math.floorDiv(index, n);
				o = index - p*n + 1;
				ret_L.add(new int[]{p+1,o});
				
				b = b ^ c;
			}
			
		}		
		return ret_L;
	}
	
	public void add(int i, int j)
	{
		bv.add((i-1)*n + (j-1));
	}
	
	public void add(int[] pair)
	{
		bv.add((pair[0]-1)*n + (pair[1]-1));
	}
	
	public void remove(int[] pair)
	{
		bv.remove((pair[0]-1)*n + (pair[1]-1));
	}
	
	public int numInt()
	{
		return bv.list.length;
	}
		
	public String toString()
	{
		return bv.toString();
	}

}
